import secrets
import asyncio
from shlex import quote
from typing import Iterable, Union

import aiofiles

from .process import run_cmd, Process
from .files import write_file
from build_django import const


def generate_env(debug: bool = False, hosts: str = '*') -> str:
    items = [
        f'SECRET_KEY={secrets.token_hex(128)}',
        f'ALLOWED_HOSTS={hosts}'
    ]

    if debug:
        items.append('DEBUG=True')

    return '\n\n'.join(items)


async def edit_settings(file) -> list[str]:
    result = []

    async for line in file:
        if line.startswith('BASE_DIR'):
            result.append('\nfrom environ import Env\n')

            result.append(f'\n{line}\n')

            result.append(f'\nenv = Env({const.DEFAULT_ENV})\n')

            result.append('\nEnv.read_env(BASE_DIR / \'.env\')\n')

        elif line.startswith('SECRET_KEY'):
            result.append('SECRET_KEY = env(\'SECRET_KEY\')\n')
        elif line.startswith('ALLOWED_HOSTS'):
            result.append('ALLOWED_HOSTS = env.tuple(\'ALLOWED_HOSTS\')')
        elif line.startswith('DEBUG'):
            result.append('DEBUG = env(\'DEBUG\')\n')
        else:
            result.append(line)

    result.extend((
        '\n# Custom settings\n'
        "\nif env('USE_SSL'):\n",
        ' ' * 4 + "SECURE_PROXY_SSL_HEADER = ('HTTP_X_FORWARDED_PROTO', 'https')"
    ))

    return result


async def write_settings(path: str):
    async with aiofiles.open(path, 'r+') as file:
        result = await edit_settings(file)

        await file.seek(0)

        await file.writelines(result)


async def write_requirements(path: str, pip_path: str):
    async with aiofiles.open(path, 'w') as file:
        await run_cmd((pip_path, 'freeze'), stdout=file)


async def install_package(pip_path: str, package: str, *args: str) -> Process:
    return await run_cmd((
        pip_path,
        'install',
        '--require-virtualenv',
        '-v',
        '--disable-pip-version-check',
        *args,
        quote(package)
    ))


async def install_packages(
    pip_path: str,
    packages: Iterable[str],
    output: Union[None, str],
    *args: str
) -> list[Process]:
    added = set()

    tasks = []

    for package in packages:
        if package in added:
            continue

        added.add(package)

        tasks.append(install_package(pip_path, package, *args))

    result = await asyncio.gather(*tasks)

    if output:
        await write_requirements(output, pip_path)

    return result



from argparse import ArgumentParser

__author__ = 'Kapustlo'

__version__ = '3.3.0'

parser: ArgumentParser = ArgumentParser(description='Build Django project')

parser.add_argument('name', metavar='name', help='Django project name')

parser.add_argument('--version', action='version', version=f'%(prog)s {__version__}')

parser.add_argument('--dir', dest='dir', type=str, default='./', required=False, help='Django project directory')

parser.add_argument('--debug', dest='debug', action='store_true', required=False, help='Should create env with DEBUG=True')

parser.add_argument('--hosts', dest='hosts', required=False, default='', help='List of comma separated ALLOWED_HOSTS values')

parser.add_argument('--python', dest='python', required=False, default=None, help='Python command. If not set, the Python used to run the program will be used')

parser.add_argument('--migrate', dest='migrate', required=False, action='store_true', help='Apply default migrations after creation')

parser.add_argument('--git', dest='git', required=False, action='store_true', help='Initialize git repo')

parser.add_argument('--commit', dest='commit', required=False, action='store_true', help='Make initial git commit')

parser.add_argument('--commit-message', dest='commit_message', required=False, default='Initial commit', help='Initial commit name')

parser.add_argument('--packages', dest='packages', nargs='+', help='Additional pip packages', required=False, default=[])

parser.add_argument('--no-compile', dest='no_compile', action='store_true', required=False, help='pip: Do not compile Python source files to bytecode')

parser.add_argument('--use-ssl', dest='us_ssl', action='store_true', required=False, help='Enable SSL support for reverse proxy')

parser.add_argument('--email', dest='email', required=False, default='', type=str, help='Git email if no global is specified')

parser.add_argument('--username', dest='username', required=False, default='', type=str, help='Git name if no global is specified')



import asyncio
from shlex import quote
from subprocess import SubprocessError
from dataclasses import dataclass
from asyncio.subprocess import Process
from typing import Iterable, Any, cast

from .shell import Shell


class GitShell(Shell):
    async def cmd(self, args, **kwargs) -> Process:
        params = self._kwargs_to_params(**kwargs)

        return await self.run(('git', *args, *params))


@dataclass
class ConfigSection(GitShell):
    section: str
    
    async def value(self, name: str) -> str:
        try:
            process = await self.cmd(('config', '--get', f'{self.section}.{name}'))
        except SubprocessError:
            return ''

        return cast(str, process.stdout)

    async def set(self, name: str, value: Any) -> Any:
        await self.cmd(('config', f'{self.section}.{name}', quote(value)))

        return self


class Config(GitShell):
    def section(self, name: str) -> ConfigSection:
        return ConfigSection(self.pwd, name)


class Repo(GitShell):
    @property
    def initialized(self) -> bool:
        return (self.pwd / '.git').exists()

    @property
    def config(self) -> Config:
        return Config(self.pwd)

    async def setup(self, email: str = '', name: str = ''):
        tasks = []

        section = self.config.section('user')

        cur_name = await section.value('name')

        if not cur_name:
            tasks.append(section.set('name', name))

        cur_email = await section.value('email')

        if not cur_email:
            tasks.append(section.set('email', email))

        await asyncio.gather(*tasks)

        return self


    async def init(self) -> Process:
        if self.initialized:
            raise ValueError(
                f'There is already a git repository at {self.pwd}'
            )

        return await self.cmd(('init',))

    async def add(self, files: Iterable[str], **kwargs) -> Process:
        args = ('add', *map(quote, files))

        return await self.cmd(args, **kwargs)

    async def commit(self, message: str, **kwargs) -> Process:
        args = ('commit', '-m', quote(message))

        return await self.cmd(args, **kwargs)



from pathlib import Path
from dataclasses import dataclass
from typing import Union, Iterable
from asyncio.subprocess import Process

from .process import run_cmd


@dataclass
class Shell:
    _pwd: Union[str, Path, None]
    
    @property
    def pwd(self) -> Path:
        return Path('./' if self._pwd is None else self._pwd).absolute()

    def _kwargs_to_params(self, **kwargs: dict[str, str]) -> list[str]:
        result = []

        for key in kwargs:
            param = key.replace('_', '-')

            value = kwargs[key]

            if isinstance(value, bool):
                result.append(param)
            else:
                result.extend((param, kwargs[key]))

        return result

    def cd(self, path: Union[str, Path]):
        path = Path(path)

        if path.is_file():
            raise ValueError(f'{path} is a file')

        self._pwd = path

        return self

    async def run(self, args: Iterable[str], **kwargs) -> Process:
        return await run_cmd(args, self.pwd, **kwargs)



import sys
import venv
import asyncio
from typing import cast
from pathlib import Path
from shlex import quote

from django.core.management import call_command

from . import const, utils, parser
from .utils.git import Repo
from .utils.files import write_file
from .utils.process import run_cmd


async def exec():
    args = parser.parse_args()

    project_path = Path(args.dir).absolute()

    print('Checking project directory...')

    if not project_path.exists():
        print('Project directory not found, creating project directory')

        project_path.mkdir()
    else:
        print('Project directory found')

    name = args.name

    print('Creating Django project...')

    call_command('startproject', name, str(project_path))

    print('Django project created')

    venv_path = project_path / 'venv'

    bin_path = venv_path / 'bin'

    pip_path = bin_path / 'pip'

    requirements_path = project_path / 'requirements.txt'

    args.packages.extend(const.REQUIRED_PACKAGES)

    print(f'Creating virtual environment at {venv_path}')

    if args.python:
        await run_cmd([
            args.python,
            '-m',
            'venv',
            str(venv_path)
        ])
    else:
        venv.create(str(venv_path), with_pip=True)

    print(f'Created virtual environment at {venv_path}')

    tasks = [
        write_file(
            str(project_path / '.gitignore'),
            const.GITIGNORE
        ),
        write_file(
            str(project_path / '.env'),
            utils.generate_env(args.debug, args.hosts)
        ),
        write_file(
            str(project_path / 'README.md'),
            const.README
        ),
        write_file(
            str(project_path / 'env.example'),
            const.ENV_EXAMPLE
        ),
        utils.write_settings(str(project_path / args.name / 'settings.py')),
        utils.install_packages(
            str(pip_path),
            args.packages,
            str(requirements_path),
            *(('--no-compile',) if args.no_compile else tuple())
        )
    ]

    repo = cast(Repo, None)  # In order to fix linter errors

    if args.git:
        repo = Repo(project_path)

        tasks.append(repo.init())

    print('Creating files...')

    await asyncio.gather(*tasks)

    print('Created files')

    tasks = []

    if args.git and args.commit:
        await repo.setup(args.email, args.username)

        await repo.add(('.',), _v=True)

        task = repo.commit(args.commit_message, _v=True)

        tasks.append(task)

    if args.migrate:
        python_path = bin_path / 'python'

        manage_path = project_path / 'manage.py'

        task = run_cmd((
            str(python_path),
            quote(str(manage_path)),
            'migrate',
            '--no-input'
        ))

        tasks.append(task)

    await asyncio.gather(*tasks)


def main():
    asyncio.run(exec())


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        sys.exit()



import asyncio
from pathlib import Path
from subprocess import SubprocessError
from asyncio.subprocess import PIPE, Process
from typing import Union, Iterable

from . import system


async def run_cmd(
    args: Iterable[str],
    cwd: Union[str, None, Path] = None,
    **kwargs
) -> Process:
    if cwd:
        cwd = str(cwd)

    cmd = ' '.join(args)

    kwargs = {
        'stdout': PIPE,
        'stderr': PIPE,
        'cwd': cwd,
        **kwargs
    }

    process = await asyncio.create_subprocess_shell(cmd, **kwargs)

    result = await process.communicate()

    out, err = result

    code = process.returncode

    if code != 0:
        raise SubprocessError(f'{code}: {err}')

    if out:
        if isinstance(out, bytes):
            out = system.decode(out)

        print(out)

    return process



GITIGNORE: str = '/venv\n/static\n/media\n__pycache__\n*.sqlite*\n*.log\n.env'

ENV_EXAMPLE: str = '''
SECRET_KEY=

# List of comma separated values
ALLOWED_HOSTS=

# Remove if production
DEBUG=

# Sets SECURE_PROXY_SSL_HEAD to ('HTTP_X_FORWARDED_PROTO', 'https')
# Remove if not needed
USE_SSL=
'''

DEFAULT_ENV = '''
    DEBUG=(bool, False),
    USE_SSL=(bool, False)
'''

REQUIRED_PACKAGES: list[str] = (
    'wheel',
    'django',
    'django-environ',
    'gunicorn'
)

README = '''
This is a Django project bootstrapped with [build-django](https://pypi.org/project/build-django/) by [Kapustlo](https://notabug.org/kapustlo)
'''
